#include "globals.hlsli"
#include "ShaderInterop_Postprocess.h"

PUSHCONSTANT(postprocess, PostProcess);

Texture2D<float4> input : register(t0);

RWTexture2D<float4> output : register(u0);

[numthreads(POSTPROCESS_BLOCKSIZE, POSTPROCESS_BLOCKSIZE, 1)]
void main(uint3 DTid : SV_DispatchThreadID)
{
	const float2 uv = (DTid.xy + 0.5f) * postprocess.resolution_rcp;
	const float amount = postprocess.params0.x;

	const float2 distortion = (uv - 0.5f) * amount * postprocess.resolution_rcp;

	const float2 uv_R = uv + float2(1, 1) * distortion;
	const float2 uv_G = uv + float2(0, 0) * distortion;
	const float2 uv_B = uv - float2(1, 1) * distortion;

	const float R = input.SampleLevel(sampler_linear_clamp, uv_R, 0).r;
	const float G = input.SampleLevel(sampler_linear_clamp, uv_G, 0).g;
	const float B = input.SampleLevel(sampler_linear_clamp, uv_B, 0).b;
	
	const float4 color = float4(R, G, B, 1);

	output[DTid.xy] = color;
}
